#include "FxCyanF.hpp"

// C++ system level
#include <cstdio>         // snprintf
// #include <cstdlib>
// #include <algorithm>
// #include <functional>

// ESP32 specific
#include "esp_log.h"

// misc
// #include <nlohmann/json.hpp>

// qthing stuff
// #include <qthing>
// #include <qthing/mqtt_common.hpp>

// project specific
#include "NVSExplorer.hpp"
#include <SpiderLib/NumberTypes.hpp>
#include <SpiderLib/ObjectTypes.hpp>



namespace SiliconTorch {

  namespace FxCyanF {


    const str ServiceNamespace = "fxCyan";


    bool configureFromNVS(FxCyanF& target) {

      if (target.getChannelCount() > 0) {
        ESP_LOGW(TAG, "Target mustn't have any channels configured");
        return false;
      }


      u8  state = 0;  // disabled -> 0   enabled -> 1
      state = NVSExplorer::NVSExplorer::instance().getUnsignedInt(ServiceNamespace, "state", state);

      if (state == 0) {
        ESP_LOGI(TAG, "Service[ FxCyanF ] disabled[]");
        return false;
      }

      ESP_LOGI(TAG, "Enabling Service[ FxCyanF ]");

      u32 frq = DefaultFrequency;
      u8  res = DefaultResolution;

      frq = NVSExplorer::NVSExplorer::instance().getUnsignedInt(ServiceNamespace, "frequency", frq);
      res = NVSExplorer::NVSExplorer::instance().getUnsignedInt(ServiceNamespace, "resolution", res);


      ESP_LOGI(TAG, "Configuring: FxCyanF{ frequency[ %d Hz ]  resolution[ %d Bit ] }", frq, res);
      target.setFrqRes(frq, res);

      for (u8 ch = 0; ch < MAX_CHANNELS; ch++) {


        char buffer[32];
        snprintf(buffer, 32, "ch%d_gpio", ch);

        u8 gpio = NVSExplorer::NVSExplorer::instance().getUnsignedInt(ServiceNamespace, buffer, 0xFF) & 0xFF;

        snprintf(buffer, 32, "ch%d_pwm", ch);
        f32 pwm = NVSExplorer::NVSExplorer::instance().getFloat(ServiceNamespace, buffer);

        // NaN == "no value stored" or any other reading error
        if (std::isnan(pwm)) pwm = 0.0f;

        if (gpio != 0xFF) {

          bool result = target.addChannel(gpio, pwm);

          // TODO: do something with result…?

          ESP_LOGI(TAG, "Configuring: FxCyanF{ channel[ %d ]  gpio[ %d ]  pwm[ %f ] }", ch, gpio, pwm);

        } else {
          break;
        }
      }

      return true;
    }


  }
}
