#pragma once

// ESP32 specific
#include "driver/ledc.h"

// C++ system level
#include <cmath>
#include <functional>

// project specific
#include "CyanBus.hpp"
#include "Metrics.hpp"
#include "PWMChannel.hpp"
#include <SpiderLib/NumberTypes.hpp>
#include <SpiderLib/ObjectTypes.hpp>



namespace SiliconTorch {

  namespace FxCyanF {

    extern const char* TAG;
    extern const str   nvsNameSpace;  // "fxCyan"
    extern const str   HEADER;        // "fxCyanF"

    constexpr u8 MAX_CHANNELS = 8;    // vorerst 8


    constexpr u32 DefaultFrequency  = 1337;  // Hz
    constexpr u8  DefaultResolution = 10;    // Bits


    using PacketHandledCallback = std::function<void()>;
    using GammaCorrector        = std::function<float(float)>;


    class FxCyanF {
      public:
        FxCyanF(u32 baseChannel = 0);

        bool addChannel(u8 gpio, float initialValue = 0.0f);

        void setPWM(u8 channel, float value);

        void setGammaCorrector(GammaCorrector gammaCorrector);

        u8 getChannelCount();

        void setBaseChannel(u16 baseChannel);
        u16 getBaseChannel();

        bool setFrequency(u32 frq_hz);
        u32 getFrequency();

        bool setResolution(u8 res_bits);
        u8 getResolution();

        bool setFrqRes(u32 frq_hz, u8 res_bits);

        void setPacketHandledCallback(PacketHandledCallback callback);

        bool handleUnicast(const u8* data, std::size_t length);
        bool handleBroadcast(const u8* data, std::size_t length);

        void registerAtCyanBus(CyanBus::CyanBus& cyanBus);

        std::string genDeviceTopic(const char* suffix);
        std::string genServiceTopic(const char* suffix);

        void publishFrqRes();
        void publishFrequency();
        void publishResolution();
        void publishBaseChannel();
        void publishChannelCount();
        void publishListenerInfo();

      private:

        u32 frequency = 1000;  // Hz
        u8 resolution =   10;  // bits

        u32 baseChannel;
        u8 channelsConfigured = 0;

        Impl::PWMChannel** channels;

        ledc_timer_config_t timer_cfg;

        GammaCorrector gammaCorrector = [](float f) { return f*f; };

        PacketHandledCallback packetCallback = []() {};
        void callPacketCallback();

        Metrics::Metrics metrics;
        Metrics::Incrementer frameCntInc;
        Metrics::Incrementer errorCntInc;
    };


    bool configureFromNVS(FxCyanF& target);

  }
}
